#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    https://gitlab.com/aspellip/twitchview

import configparser
import tkinter as tk
from io import BytesIO
import json
import os
import random
import subprocess
import sys
import threading
import traceback
import urllib.request
import urllib.error
import urllib.parse
from functools import partial

import m3u8
from PIL import Image, ImageTk

from twitchview.irc import main as ircchat

__version__ = "0.6.8"


class AppTk(tk.Tk):
    def __init__(self, parent, config):
        tk.Tk.__init__(self, parent)
        self.parent = parent
        self.cfg = configparser.ConfigParser()
        self.cfg.read(config + 'config.ini')
        self.config_path = config
        # This token you can get on a URL http://tw.fex.cc
        self.token = self.cfg.get('twitch', 'token')
        self.nick = self.cfg.get('twitch', 'nick')
        self.debug = self.cfg.get('twitch', 'debug')
        self.player = self.cfg.get('twitch', 'player')
        self.tw_url = "https://api.twitch.tv/kraken/streams/followed"
        self.usher_api = 'https://usher.ttvnw.net/api/channel/hls/{channel}.m3u8?player=twitchweb' + \
                         '&token={token}&sig={sig}&$allow_audio_only=true&allow_source=true' + \
                         '&type=any&p={random}'
        self.token_api = 'https://api.twitch.tv/api/channels/{channel}/access_token?oauth_token='
        self.logo_size = 32, 32
        self.create_menubar()
        self.initialize()

    def print_debug(self, message):
        if self.debug == "on":
            print(message)

    @staticmethod
    def url_request(url, token):
        data = {"Authorization": "OAuth %s" % token, 'Accept': 'application/vnd.twitchtv.v5+json'}
        req = urllib.request.Request(url, headers=data)
        response = urllib.request.urlopen(req)
        return response

    def get_token_and_signature(self, channel):
        try:
            url = self.token_api.format(channel=channel)
            response = urllib.request.urlopen('%s%s' % (url, self.token))
            data = json.load(response)
        except urllib.error.HTTPError as e:
            self.print_debug(e)
            return False
        sig = data['sig']
        token = urllib.parse.quote(data['token'])
        return token, sig

    def get_live_stream(self, channel):
        token, sig = self.get_token_and_signature(channel)
        r = random.randint(0, 1E7)
        try:
            url = self.usher_api.format(channel=channel, sig=sig, token=token, random=r)
            response = urllib.request.urlopen(url)

        except urllib.error.URLError as e:
            self.print_debug(e)
            return False
        data = response.read().decode("utf-8")

        m3u8_obj = m3u8.loads(data)
        return m3u8_obj

    @staticmethod
    def print_video_urls(m3u8_obj):
        quality = {}
        try:
            for p in m3u8_obj.playlists:
                quality[p.media[0].name.split(' ', 1)[0]] = p.uri
        except KeyError:
            traceback.print_exc()
            sys.exit(3)
        return quality

    def create_menubar(self):
        menubar = tk.Menu(self)
        filemenu = tk.Menu(self, menubar, tearoff=0)
        filemenu.add_command(label="Reload", command=self.restart_ui, accelerator="Ctrl+Shift+R")

        filemenu.add_separator()

        filemenu.add_command(label="Exit", command=self.destroy_ui, accelerator="Ctrl+Shift+Q")
        menubar.add_cascade(label="File", menu=filemenu)

        helpmenu = tk.Menu(menubar, tearoff=0)
        helpmenu.add_command(label="About...", command=self.create_window_about)
        menubar.add_cascade(label="Help", menu=helpmenu)

        self.config(menu=menubar)
        self.bind_all("<Control-R>", self.restart_ui)
        self.bind_all("<Control-Q>", self.destroy_ui)

    def get_twitch_favorite(self):
        try:
            response = self.url_request(self.tw_url, self.token)
            data = json.load(response)
            return data
        except urllib.error.HTTPError as e:
            self.print_debug(e)
            return None

    def worker(self, i, j):
        # Get logo images of streamers and resize them
        img_data = BytesIO(urllib.request.urlopen(i['channel']['logo']).read())
        image = Image.open(img_data)
        image.thumbnail(self.logo_size, Image.ANTIALIAS)
        photo = ImageTk.PhotoImage(image.convert("RGB"))

        tw_col = {
            'status': i['channel']['status'],
            'name': i['channel']['name'],
            'game': i['game'],
            'viewers': i['viewers']
        }
        # Logo is a first in a row.
        label_img = tk.Label(self, image=photo, justify=tk.LEFT)
        label_img.image = photo
        label_img.grid(row=j, column=0)

        for k, col in enumerate(tw_col.values()):
            try:
                col = str(col)
                self.print_debug(col)
                tk.Label(self, text=col[:64], justify=tk.LEFT).grid(row=j, column=k + 1, sticky=tk.W)
            except tk.TclError:
                tk.Label(self, text=":)", justify=tk.LEFT).grid(row=j, column=k + 1, sticky=tk.W)

        # Button is a last in a row
        self.print_debug(tw_col.values())
        m3u8_obj = self.get_live_stream(tw_col['name'])
        
        q_buttons = self.print_video_urls(m3u8_obj)
        v = tk.StringVar(self)
        v.set(list(q_buttons.keys())[0])
        tk.OptionMenu(self, v, *q_buttons.keys()).grid(row=j, column=len(tw_col) + 2 + 1, sticky=tk.W)
        tk.Button(self, text="Play", justify=tk.LEFT,
                  command=partial(self.button_play_stream, q_buttons, v)).grid(row=j, column=len(tw_col) + 4,
                                                                               sticky=tk.W)
        tk.Button(self, text="Chat", justify=tk.LEFT,
                  command=partial(self.button_chat, tw_col['name'])).grid(row=j, column=len(tw_col) + 5, sticky=tk.W)

    def initialize(self):
        streamers = self.get_twitch_favorite()
        try:
            if streamers['streams'] is not None:
                threads = []
                for j, i in enumerate(streamers['streams']):
                    t = threading.Thread(target=self.worker, args=(i, j,))
                    threads.append(t)
                    t.start()
            else:
                tk.Label(self, text="No channels available in list",
                         justify=tk.LEFT).grid()
        except TypeError:
            tk.Label(self, text="Wrong token in %sconfig.ini OR twitch servers unavailable" % self.config_path,
                     justify=tk.LEFT).grid()

    def button_play_stream(self, uri, quality):
        quality = quality.get()
        uri = uri[quality]
        self.print_debug(quality)
        if "nt" == os.name:
            startupinfo = subprocess.STARTUPINFO()
            startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
            subprocess.Popen(
                [self.player, uri,
                 quality], startupinfo=startupinfo,
                stderr=subprocess.PIPE,
                stdout=subprocess.PIPE)
        else:
            subprocess.Popen(
                [self.player, uri,
                 quality], stderr=subprocess.PIPE,
                stdout=subprocess.PIPE)

    def button_chat(self, tw_name):
        ircchat(self.nick, self.token, "#%s" % tw_name)

    def create_window_about(self):
        t = tk.Toplevel(self)
        t.wm_title("About...")
        lw = tk.Label(t, text="This is twitchview v%s\n src: https://gitlab.com/aspellip/twitchview" % __version__)
        lw.pack(side="top", fill="both", expand=True, padx=100, pady=100)

    def restart_ui(self, event=""):
        self.print_debug("Restarting .... %s" % event)
        self.destroy()
        main()

    def destroy_ui(self, event=""):
        self.print_debug("Quit .... %s" % event)
        self.destroy()


def check_config():
    home_dir = os.environ['HOME'] + "/.twitchview/"
    config_file = home_dir + "config.ini"
    if not os.path.exists(home_dir):
        os.makedirs(home_dir)
    if not os.path.exists(config_file):
        from shutil import copyfile
        copyfile("config.ini.example", config_file)
    return home_dir


def main():
    os.chdir(os.path.dirname(os.path.realpath(__file__)))
    config = check_config()
    app = AppTk(None, config)
    app.title('Twitch View')
    if "nt" == os.name:
        app.wm_iconbitmap(bitmap="icon.ico")
    else:
        app.wm_iconbitmap(bitmap="@icon.xbm")
    app.mainloop()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) 2016 Akuli

# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""Simple IRC client written in Tkinter.

Mostly based on this:
  http://www.ohjelmointiputka.net/koodivinkit/24802-python-ircbot

TODO:
- Add /me, /msg, /join and /part, but not support for multiple channels.
  This is a minimal client.
"""

import collections
import socket
import threading
import time
import tkinter as tk
import sys


class ClientCore:
    """The core of the IRC client."""

    def __init__(self, server, nick, password, channel, port=6667, outfunc=None, logfile=None):
        """Initialize the client.

        The username and realname keyword arguments default to nick, and
        outfunc defaults to print.
        """
        self._server = server
        self._nick = nick
        self._channel = channel
        self._port = port
        self._password = password
        self._outfunc = print if outfunc is None else outfunc
        self._logfile = None if logfile is None else open(logfile, 'a')

    def _send(self, msg):
        """Send a message.

        Encode the message, add a line end and send it to the server.
        """
        self._socket.send(msg.encode('utf-8', errors='replace') + b'\r\n')

    def send_to_channel(self, msg):
        """Send a message to channel if it's non-empty."""
        if msg:
            self.show(self._nick, msg)
            self._send('PRIVMSG {} :{}'.format(self._channel, msg))

    def connect(self):
        """Connect the client.

        This will connect the socket, set nickname, username and
        realname and join a channel.
        """
        self._socket = socket.socket()
        self._socket.connect((self._server, self._port))
        self._send('PASS {}'.format(self._password))
        self._send('NICK {}'.format(self._nick))
        self._send('JOIN {}'.format(self._channel))

    def _check(self, line):
        try:
            beginning, msg = line.lstrip(':').split(':', 1)
            sender, msg_type, target = beginning.split()
        except ValueError:
            # Not a message.
            if line.startswith('PING '):
                self._send('PONG :abc')
        else:
            self.show(sender.split('!')[0], msg)

    def show(self, sender, msg):
        """Show a message to the user.

        Call self._outfunc and write to the log.
        """
        self._outfunc(sender, msg)
        # if self._logfile is not None:
        #    print(self.format_msg(sender, msg), file=self._logfile, flush=True)

    def outputloop(self):
        """Receive data from the channel and write it to outstream."""
        try:
            while True:
                try:
                    buf = b''
                    while True:
                        buf += self._socket.recv(4096)
                        *lines, buf = buf.split(b'\r\n')
                        for line in lines:
                            self._check(line.decode('utf-8', errors='replace'))
                except socket.error as e:
                    self.show("*", "Error: {}".format(e))
                    self.show("*", "Trying again...")
                    self.connect()
        finally:
            if self._logfile is not None:
                self._logfile.close()

    @staticmethod
    def format_msg(sender, msg):
        """Return a printable form of the message."""
        return '[{}] {:>20} | {}'.format(time.strftime('%H:%M:%S'),
                                         sender, str(msg))


class ClientGUI(tk.Tk):
    """A GUI for ClientCore."""

    def __init__(self):
        """Initialize the GUI."""
        tk.Tk.__init__(self)

        textarea = tk.Frame(self)
        self._text = tk.Text(textarea, state='disabled')
        self._text.pack(side='left', fill='both', expand=True)
        scrollbar = tk.Scrollbar(textarea, command=self._text.yview)
        scrollbar.pack(side='right', fill='y')
        self._text['yscrollcommand'] = scrollbar.set
        textarea.pack(fill='both', expand=True)

        entry = tk.Entry(self, font='TkFixedFont')
        entry.pack(fill='x')
        entry.bind('<Return>', self._on_enter)
        entry.bind('<Control-A>', self._on_control_a)
        entry.bind('<Control-a>', self._on_control_a)

        self.bind_all('<Control-Q>', lambda e: self.destroy())
        self.bind_all('<Control-q>', lambda e: self.destroy())

        # New messages are stored here instead of adding them directly
        # to the window. The core is ran in another thread and it
        # doesn't access tkinter this way.
        self._msg_queue = collections.deque()

    def create_core(self, **core_args):
        """Create and connect a core.

        This can be called only once.
        """
        if hasattr(self, '_core'):
            raise RuntimeError("cannot create core twice")
        self._core = ClientCore(outfunc=self._queue_msg, **core_args)
        self._core.connect()
        theadIRC = threading.Thread(target=self._core.outputloop)
        theadIRC.daemon = True
        theadIRC.start()
        self._clear_queue_loop()

    def _queue_msg(self, sender, msg):
        """Add a message to the message queue."""
        self._msg_queue.append(self._core.format_msg(sender, msg))

    def _clear_queue_loop(self):
        """Show each message in the queue."""
        if self._msg_queue:
            self._text['state'] = 'normal'
            while True:
                try:
                    self._text.insert('end', self._msg_queue.popleft())
                    self._text.insert('end', '\n')
                    self._text.see('end')
                except (IndexError, tk.TclError):
                    break
            self._text['state'] = 'disabled'
        self.after(100, self._clear_queue_loop)

    def _on_enter(self, event):
        """Send a message to the channel."""
        entry = event.widget
        msg = entry.get()
        self._core.send_to_channel(msg)
        entry.delete(0, 'end')

    def _on_control_a(self, event):
        """Select all in the entry."""
        entry = event.widget
        entry.selection_range(0, 'end')
        return 'break'


def main(nick, password, channel):
    core_args = {
        'server': "irc.twitch.tv",
        'nick': nick,
        'password': "oauth:%s" % password,
        'channel': channel,
        'port': 6667,
    }
    root = ClientGUI()
    root.title(channel)
    root.create_core(**core_args)
    root.mainloop()



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
setup.py for the Python packager

Based on https://github.com/pypa/sampleproject/blob/master/setup.py
"""

# Always prefer setuptools over distutils
from setuptools import setup
# To use a consistent encoding
from codecs import open
from os import path
import os
from twitchview import main

here = path.abspath(path.dirname(__file__))
print(here)
try:
    README = open(os.path.join(here, 'README.rst')).read()
except IOError:
    README = ''
try:
    TODO = open(os.path.join(here, 'TODO.txt')).read()
except IOError:
    TODO = ''
try:
    AUTHORS = open(os.path.join(here, 'CONTRIBUTING.amd')).read()
except IOError:
    AUTHORS = ''
try:
    CHANGES = open(os.path.join(here, 'CHANGELOG')).read()
except IOError:
    CHANGES = ''

long_description = README + '\n\n' + TODO + '\n\n' + AUTHORS + '\n\n' + CHANGES

print(os.listdir('.'))  # DEBUG

setup(
    name='twitchview',
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=main.__version__,
    description='Simple program by python for watching favorite twitch channels',
    long_description=long_description,
    url='https://gitlab.com/aspellip/twitchview',
    bugtrack_url='https://gitlab.com/aspellip/twitchview/issues',
    author='Aspel',
    author_email='aspel@ukr.net',
    license='GPLv3',
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        "Operating System :: POSIX",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        'Programming Language :: Python :: 3',
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Multimedia :: Sound/Audio",
        "Topic :: Multimedia :: Video",
        "Topic :: Utilities"
    ],
    keywords='twitch',
    packages=['twitchview'],
    entry_points={
          "gui_scripts": ["twitchview=twitchview.main:main"]
      },
    package_data={
        # If any package contains *.txt or *.rst files, include them:
        '': ['*.ico', '*.xbm', '*.example', '*.rst', '*.txt', '*.pyw'],
        # And include any *.msg files found in the 'hello' package, too:
        'hello': ['*.msg'],
    },
    python_requires=">=3",
    install_requires=['m3u8', 'Pillow']
    # Actually also requires numpy, scipy and numpy but I don't want to force
    # pip to install these since pip is bad at that for those packages.
)
